/*

		MultiMediaCard Driver Module
		for FOX 2.6 kernel

		Simple single card bit-bang MMC/SD driver

		(c) 2007 AFM Design - www.afmdesign.net

	License:
	    This program is free software: you can redistribute it and/or modify	
	    it under the terms of the GNU General Public License as published by
	    the Free Software Foundation, either version 3 of the License, or
	    (at your option) any later version.

	    This program is distributed in the hope that it will be useful,
	    but WITHOUT ANY WARRANTY; without even the implied warranty of
	    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	    GNU General Public License for more details.
	
	    You should have received a copy of the GNU General Public License
	    along with this program.  If not, see <http://www.gnu.org/licenses/>.

	Notes:
		Basic hardware connections as described on Acme's site www.acmesystems.it
		Additional support for insert detect and LED, defined in mmc.h
		Go to www.afmdesign.net for more info

		Assumes card sectorsize always 512 bytes

		Reads/writes in stream mode, idle timer typically flushes after 5 sec and
		puts card into standby mode.
		
		Removing a mounted card can cause data corruption.

		Identifies 'legacy' SD cards as MMC,
		High Capacity SD as either HCSD or SD.

		Tested with SanDisk 	256 MB (SD), 1 GB (SD), 4 GB (HCSD, block mode)
					DaneElec	1 GB (SD), 2 GB (SD, fake 1024 byte sectorsize)

	Credits:
		This is a complete redesign of the MMC/SD driver for the Acme FOX,
		inspired by K. John '2B|!2B' Crispin's version.

	Version 1.01	3 Oct 2007	by AFM
		New release
*/


// Includes

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/blkdev.h>
#include <linux/genhd.h>

#include "mmc.h"
#include "mmc_def.h"
#include "mmc_io.h"
#include "mmc_ctrl.h"
#include "mmc_spi.h"

MODULE_LICENSE("GPL");
MODULE_AUTHOR("AFM");
MODULE_DESCRIPTION("AFM MMC Driver");


// Defines

#define MMC_VERSION				"V1.01"

#define MMC_MAJOR				126
#define MMC_DEVICE				"mmc"
#define MMC_IDLE_TIME			5


// Globals

static struct gendisk *mmc_gendisk;
static spinlock_t mmc_spinlock;
static struct request_queue *mmc_queue;
static struct timer_list mmc_timer;
static unsigned char mmc_disk_active;
static unsigned char mmc_card_changed;
static unsigned char mmc_media_ok;
static unsigned char mmc_idle_timer;
static int mmc_disk_capacity;


// Module functions

void mmc_request(request_queue_t *q)
{
	struct request *req;
	int r;
	int success = 1;

	while ((req = elv_next_request(q)) != NULL) {
		if (! blk_fs_request(req)) {
			end_request(req, 0);
			continue;
		}
		mmc_idle_timer = MMC_IDLE_TIME;
		spin_unlock_irq(q->queue_lock);
		if (mmc_disk_active && MMC_CARD_PRESENT) {
			if (rq_data_dir(req) == READ) {
	            if ((r = mmc_read_sectors(req->buffer, req->sector, req->current_nr_sectors)) != req->current_nr_sectors) {
					printk(KERN_ERR "mmc: Read sector (%lu, %u) failed (%u)\n", req->sector, req->current_nr_sectors, r);
					success = 0;
				}
			}
			if (rq_data_dir(req) == WRITE) {
	            if ((r = mmc_write_sectors(req->buffer, req->sector, req->current_nr_sectors)) != req->current_nr_sectors) {
					printk(KERN_ERR "mmc: Write sector (%lu, %u) failed (%u)\n", req->sector, req->current_nr_sectors, r);
					success = 0;
				}
			}

		} else printk(KERN_ERR "mmc: Request failed, disk not active/present\n");

		spin_lock_irq(q->queue_lock);
		end_request(req, success);
	}
}

int mmc_open(struct inode *inode, struct file *filp)
{
	if (MMC_CARD_PRESENT) {
		if ((mmc_disk_capacity = mmc_init_card()) == -1) {
			mmc_media_ok = 0;
			printk(KERN_ERR "mmc: Media init failed\n");
		} else {
			mmc_media_ok = 1;
			mmc_card_changed = 1;
			check_disk_change(inode->i_bdev);
			mmc_led_on();
			return 0;
		}
	} else printk(KERN_ERR "mmc: Card not present\n");

	return -ENODEV;
}

int mmc_release(struct inode *inode, struct file *filp)
{
	mmc_idle();
	mmc_led_off();
	mmc_disk_active = 0;
	set_capacity(mmc_gendisk, 0);

	return 0;
}

int mmc_revalidate(struct gendisk *disk)
{
	if (mmc_media_ok && MMC_CARD_PRESENT) {
		set_capacity(mmc_gendisk, mmc_disk_capacity);
		mmc_disk_active = 1;
	}
	mmc_card_changed = 0;
	return 0;
}

int mmc_check_media_changed(struct gendisk *dev)
{
	return (mmc_card_changed);
}

static struct block_device_operations mmc_ops = {
	.open = mmc_open,
	.release = mmc_release,
	.media_changed = mmc_check_media_changed,
	.revalidate_disk = mmc_revalidate,
	.owner = THIS_MODULE
};

int mmc_init_disk(void)
{
	spin_lock_init(&mmc_spinlock);
	mmc_queue = blk_init_queue(mmc_request, &mmc_spinlock);
    if (mmc_queue == NULL) {
		printk(KERN_INFO "mmc: Queue init failed\n");
		return -1;	
	}
	if (!(mmc_gendisk = alloc_disk(5))) {
		printk(KERN_INFO "mmc: alloc_disk failed\n");
		blk_cleanup_queue(mmc_gendisk->queue);
		return -1;	
	}
	mmc_gendisk->queue = mmc_queue;	
	mmc_gendisk->major = MMC_MAJOR;
   	mmc_gendisk->first_minor = 1;
   	mmc_gendisk->fops = &mmc_ops;
   	strcpy (mmc_gendisk->disk_name, MMC_DEVICE);
	set_capacity(mmc_gendisk, 0);
	add_disk(mmc_gendisk);

	return 0;
}


void mmc_timer_event(unsigned long arg)
{
	static int state = 0;

	switch (state) {
		case 0:
			if (MMC_CARD_PRESENT) state = 1;
			break;

		case 1:
			if (MMC_CARD_PRESENT) {
				if (MMC_INSERT_DETECT) printk(KERN_INFO "mmc: Card inserted\n");
				mmc_idle_timer = 0;
				state = 2;
			} else state = 0;
			break;

		case 2:
			if (MMC_CARD_PRESENT) {
				if (mmc_idle_timer && mmc_disk_active)
					if (--mmc_idle_timer == 0) mmc_idle();
			} else {
				if (mmc_disk_active) printk(KERN_CRIT "mmc: Card removed in open state, possible data corruption\n");
				else printk(KERN_INFO "mmc: Card removed\n");
				mmc_disk_active = 0;
				state = 0;
			}
			break;
	}
	mod_timer(&mmc_timer, jiffies + HZ);
}


static int __init mmc_init(void)
{
	printk("mmc: MultiMediaCard Driver %s - (c) 2007 AFM Design\n", MMC_VERSION);

	if (register_blkdev(MMC_MAJOR, MMC_DEVICE) < 0) {
		printk(KERN_ERR "mmc: Device registration failed\n");
		return -1;
	}

	mmc_spi_setup();

	mmc_disk_active = 0;
	mmc_card_changed = 0;
	mmc_media_ok = 0;

	init_timer(&mmc_timer);
	mmc_timer.function = mmc_timer_event;
	mmc_timer.expires = jiffies + HZ;
	add_timer(&mmc_timer);	

	mmc_init_disk();

	if (!MMC_INSERT_DETECT) printk(KERN_INFO "mmc: Card insert detect disabled\n");
	if (!MMC_LED_ENABLE) printk(KERN_INFO "mmc: Activity LED disabled\n");

	printk(KERN_INFO "mmc: Device driver initialized\n");

	return 0;
}

static void __exit mmc_exit(void)
{
	del_timer(&mmc_timer);
	blk_cleanup_queue(mmc_gendisk->queue);
	del_gendisk(mmc_gendisk);
	put_disk(mmc_gendisk);
	unregister_blkdev(MMC_MAJOR, MMC_DEVICE);
	printk("mmc: Module unloaded\n");
}


module_init(mmc_init);
module_exit(mmc_exit);
