/*

		MultiMediaCard Driver Module
		for Fox 2.6 kernel

		MMC Controller functions


		(c) 2007 AFM Design - www.afmdesign.net

	License:
	    This program is free software: you can redistribute it and/or modify	
	    it under the terms of the GNU General Public License as published by
	    the Free Software Foundation, either version 3 of the License, or
	    (at your option) any later version.

	    This program is distributed in the hope that it will be useful,
	    but WITHOUT ANY WARRANTY; without even the implied warranty of
	    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	    GNU General Public License for more details.
	
	    You should have received a copy of the GNU General Public License
	    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

// Includes

#include <linux/kernel.h>
#include <asm/io.h>

#include "mmc_def.h"
#include "mmc_spi.h"


// Defines

// MMC Mode
#define MMC_MODE_SINGLE				0
#define MMC_MODE_READ				1	
#define MMC_MODE_WRITE				2


// Globals

static unsigned char mmc_mode = MMC_MODE_SINGLE;


// Functions

unsigned char mmc_crc7(unsigned char *buf, int len)
{
	int i, j;
	unsigned char c, crc;

	crc = 0;
	for (i = 0; i < len; i++) {
		c = *buf++;
		for (j = 0; j < 8; j++) {
			crc <<= 1;
			if ((c ^ crc) & 0x80) crc ^= 0x09;
			c <<= 1;
		}
	}
	crc <<= 1;
	crc |= 0x01;
	return (crc);
}

int mmc_check_busy(void)
{
	int i;

	mmc_spi_clk(1);
	for (i = 0; i < MMC_MAX_BUSY_LOOPS; i++) {
		if (mmc_spi_read() == 0xFF) {
			mmc_spi_clk(1);
			return 0;
		}
	}

	printk(KERN_WARNING "mmc: check_busy loops exceeded\n");
	mmc_spi_clk(1);
	return -1;
}

int mmc_get_token(void)
{
	int i;
	unsigned char t;

	for (i = 0; i < MMC_MAX_TOKEN_LOOPS; i++) {
		t = mmc_spi_read();
		if (t != 0xFF) {
			return (int)t;
		}
	}

	printk(KERN_WARNING "mmc: get_token loops exceeded\n");
	return -1;
}


int mmc_send_cmd_r1(unsigned char cmd, unsigned int arg)
{
	unsigned char r1;
	int i;
	unsigned char buf[6];

	mmc_spi_clk(8);

	buf[0] = cmd | 0x40;
	buf[1] = (unsigned char)((arg >> 24) & 0xFF);
	buf[2] = (unsigned char)((arg >> 16) & 0xFF);
	buf[3] = (unsigned char)((arg >> 8) & 0xFF);
	buf[4] = (unsigned char)((arg) & 0xFF);
	buf[5] = mmc_crc7(buf, 5);

	mmc_spi_write_n(buf, sizeof(buf));
	mmc_spi_clk(1);

	for (i = 0; i < 8; i++) {
		r1 = mmc_spi_read();
		if ((r1 & 0x80) == 0) {
			return (int)r1;
		}
	}
	printk(KERN_WARNING "mmc: send_cmd_r1 didn't respond\n");
	return -1;
}

int mmc_idle(void)
{
	MMC_SELECT;
	if (mmc_check_busy() == -1) return -1;
	if (mmc_mode == MMC_MODE_READ) {
		if (mmc_send_cmd_r1(MMC_STOP_TRANSMISSION, 0) == -1) return -1; 
	}
	if (mmc_mode == MMC_MODE_WRITE) {
		mmc_spi_write(MMC_STOP_TRAN);
		if (mmc_check_busy() == -1) return -1;
	}
	mmc_mode = MMC_MODE_SINGLE;
	MMC_DESELECT;
	return 0;
}

int mmc_cmd(unsigned char cmd, unsigned int arg)
{
	int r1;

	MMC_SELECT;
	r1 = mmc_send_cmd_r1(cmd, arg); 
	MMC_DESELECT;

	return r1;
}

int mmc_cmd_read(unsigned char cmd, unsigned int arg, unsigned char *buf, unsigned short len)
{
	int r1;

	MMC_SELECT;
	r1 = mmc_send_cmd_r1(cmd, arg); 
	if ((r1 & 0x80) == 0) mmc_spi_read_n(buf, len);
	MMC_DESELECT;

	return r1;
}

int mmc_cmd_read_reg(unsigned char cmd, unsigned int arg, unsigned char *buf, unsigned short len)
{
	int r1;
	unsigned short crc;

	MMC_SELECT;
	r1 = mmc_send_cmd_r1(cmd, arg); 
	if ((r1 & 0x80) == 0) {
		if ((r1 = mmc_get_token()) == MMC_START_BLOCK) {
			mmc_spi_read_n(buf, len);
			mmc_spi_read_n(&crc, sizeof(crc));
		} else {
			printk(KERN_WARNING "mmc: read_reg unexpected token = %d\n", r1);
			r1 = -1;
		}
	}
	MMC_DESELECT;

	return r1;
}

int mmc_read_block(unsigned char *buf, unsigned int arg)
{
	static unsigned int last_arg;
	unsigned char token;
	unsigned short crc;
	int r;

	MMC_SELECT;

	if (mmc_mode == MMC_MODE_READ) {
		if (arg != (last_arg + MMC_KERNEL_SECTORSIZE)) {
			if (mmc_idle() == -1) return -1;
		}
	} 

	if (mmc_mode == MMC_MODE_WRITE) {
		if (mmc_idle() == -1) return -1;
	}

	if (mmc_mode == MMC_MODE_SINGLE) {
		MMC_SELECT;
		r = mmc_send_cmd_r1(MMC_READ_MULTIPLE_BLOCK, arg);
		if (r != 0) {
			printk(KERN_WARNING "mmc: Read multiple R1 token = %d\n", r);
			token = mmc_spi_read();
			printk(KERN_WARNING "mmc: Read data error token = %u\n", token);
			return -1;
		}
		mmc_mode = MMC_MODE_READ;
	}

	if ((r = mmc_get_token()) == MMC_START_BLOCK) {
		mmc_spi_read_n(buf, MMC_KERNEL_SECTORSIZE);
		mmc_spi_read_n(&crc, sizeof(crc));
		last_arg = arg;
	} else {
		printk(KERN_WARNING "mmc: Read data unexpected token = %d\n", r);
		r = -1;
	}

	return r;
}

int mmc_write_block(unsigned char *buffer, unsigned int arg)
{
	static unsigned int last_arg;
	unsigned char token;
	unsigned short crc = 0;
	int r = 0;

	MMC_SELECT;

	if (mmc_check_busy() == -1) return -1;

	if (mmc_mode == MMC_MODE_WRITE) {
		if (arg != (last_arg + MMC_KERNEL_SECTORSIZE)) {
			if (mmc_idle() == -1) return -1;
		}
	} 

	if (mmc_mode == MMC_MODE_READ) {
		if (mmc_idle() == -1) return -1;
	}

	if (mmc_mode == MMC_MODE_SINGLE) {
		MMC_SELECT;
		r = mmc_send_cmd_r1(MMC_WRITE_MULTIPLE_BLOCK, arg);
		if (r != 0) {
			printk(KERN_WARNING "mmc: Write multiple R1 token = %d\n", r);
			return -1;
		}
		mmc_mode = MMC_MODE_WRITE;
	}

	mmc_spi_write(MMC_START_BLOCK_2);
	mmc_spi_write_n(buffer, MMC_KERNEL_SECTORSIZE);
	mmc_spi_write_n(&crc, sizeof(crc));
	mmc_spi_read_n(&token, 1);
	if ((token & 0x1F) != MMC_DATA_ACCEPTED) {
		printk(KERN_WARNING "mmc: Write data unexpected token = %u\n", token);
		r = -1;
	} else last_arg = arg;

//	mmc_io_clk(1);

	return r;
}
